/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gpdef.h>
#include <gpstdlib.h>
#include <gpgraphic.h>
#include <gpstdio.h>
#include <gpfont.h>
#include <gpmm.h>

#include "m6809.h"
#include "pia.h"
#include "sound.h"
#include "video.h"
#include "types.h"

#define SAMPLE_RATE 22418
#define CHANNELS 1
#define FRAME_SIZE 512
#define BUFFER_SIZE (FRAME_SIZE * 2)
#define SAMPLE_CYCLES (OSCILLATOR_RATE / SAMPLE_RATE)
//#define BUFFER_CYCLES ((OSCILLATOR_RATE * BUFFER_SIZE) / SAMPLE_RATE)
#define BUFFER_CYCLES ((OSCILLATOR_RATE / SAMPLE_RATE) * BUFFER_SIZE)
#define FRAME_CYCLES ((OSCILLATOR_RATE / SAMPLE_RATE) * FRAME_SIZE)

static Uint32 sound_cycle_base;
static Uint8 *buffer;
static Uint8 *buffer_frame1;
static Uint8 *buffer_end;
static Uint8 *wrptr;
static Uint8 lastsample;
static volatile int idx_buf;
static volatile Uint32 *playpos;

int sound_init(void) {
	GpPcmInit(PCM_M22, PCM_8BIT);
	PIA_1A.wrfunc[0] = sound_write_byte;
	buffer = malloc(BUFFER_SIZE);
	memset(buffer, 0x80, BUFFER_SIZE);
	buffer_frame1 = buffer + FRAME_SIZE;
	buffer_end = buffer + BUFFER_SIZE;
	wrptr = buffer;
	lastsample = 0x80;
	GpPcmPlay((Uint16 *)buffer, BUFFER_SIZE, 1);
	GpPcmLock((Uint16 *)buffer, (int *)&idx_buf, (Uint32 *)&playpos);
	return 1;
}

void sound_reset(void) {
}

void sound_write_byte(Uint32 octet) {
	Uint32 elapsed_cycles = current_cycle - sound_cycle_base;
	Uint32 fill_with;
	Uint8 *fill_to;
	if (elapsed_cycles > BUFFER_CYCLES * 2) {
		sound_cycle_base = current_cycle;
		elapsed_cycles = 0;
	}
	fill_to = buffer + (elapsed_cycles/(Uint32)SAMPLE_CYCLES);
	fill_with = lastsample;
	octet >>= 1;
	if (!(PIA_1B.control_register & 0x08) || PIA_0B.control_register & 0x08 || PIA_0A.control_register & 0x08)
		octet = lastsample; // 0x80;
	if (wrptr < buffer_frame1) {
		while (*playpos < (Uint32)buffer_frame1);
		while (wrptr < fill_to && wrptr < buffer_frame1)
			*(wrptr++) = lastsample;
		if (wrptr >= buffer_frame1) {
			while (*playpos >= (Uint32)buffer_frame1);
			while (wrptr < fill_to)
				*(wrptr++) = lastsample;
		}
	} else {
		while (*playpos >= (Uint32)buffer_frame1);
		while (wrptr < fill_to && wrptr < buffer_end)
			*(wrptr++) = lastsample;
		if (wrptr >= buffer_end) {
			while (*playpos < (Uint32)buffer_frame1);
			fill_to -= BUFFER_SIZE;
			sound_cycle_base += BUFFER_CYCLES;
			wrptr = buffer;
			while (wrptr < fill_to)
				*(wrptr++) = lastsample;
		}
	}
	lastsample = octet;
	return;
}
